#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <libse.h>
#include <internal/exstring.h>

static char* strdup(const char* str)
{
	register size_t size = strlen(str);
	char* ret = malloc(size+1);
	memcpy(ret, str, size);
	ret[size] = 0;
	return ret;
}

static void _addcar(list_t* out, exString* token, int nil, int ucase)
{
	if(ucase)
		for(char* t = *token; *t; t++)
			*t = toupper((unsigned char)*t);
	if(nil&&strcmp(EXS(token), "NIL")==0)
		out->ptr = NULL;
	else {
		out->ptr = strdup(EXS(token));
		out->car.size = strlen(out->ptr);
		out->car.alloced = 1;
		exs_reset(token);
	}
}

static list_t* _parse_list(const char* sexpr, int* lp, unsigned int flags)
{
	list_t* out = malloc(sizeof(list_t));
	memset(out,0,sizeof(list_t));
	exString *token = exs_new(NULL);

	int instr=0;
	int inesc=0;

	while(*sexpr) {
		if(inesc)
		{
			exs_nappend(token, sexpr, 1);
			inesc=0;
		}
		else if(instr)
		{
			if(*sexpr=='\"')
				instr=0;
			else exs_nappend(token, sexpr, 1);
		}
		else if(*sexpr==' ' || *sexpr == '\r' || *sexpr == '\n' || *sexpr == '\t') {
			if(strlen(EXS(token))>0) {
				_addcar(out, token, flags & SET_PARSE_NIL, flags & SET_PARSE_INSENSITIVE);
				
				if(*(sexpr+1))
					out->cdr = _parse_list(sexpr, lp, flags);
				else out->cdr =NULL;
				break;
			}
		}
		else if(*sexpr == ')')
		{
			if(**token)
				_addcar(out, token, flags & SET_PARSE_NIL, flags & SET_PARSE_INSENSITIVE);
			out->cdr=NULL;	
			(*lp)+=1;
			break;
		}
		else if(*sexpr == '(')
		{
			if(**token) {
				_addcar(out, token, flags & SET_PARSE_NIL, flags&SET_PARSE_INSENSITIVE);
				if(*(sexpr+1))
					out->cdr = _parse_list(sexpr, lp ,flags);
				else out->cdr = NULL; //Invalid syntax.
				break;
			}
			else {
				int hp =0;
				list_t* nc = _parse_list(sexpr+1, &hp, flags);
				out->ptr = nc;
				out->car.alloced=1;
				out->car.sexpr=1;
				out->car.size = sizeof(list_t);
				sexpr+= hp+1;
				(*lp) += hp+1;
		
				if(*(sexpr))
					out->cdr = _parse_list(sexpr, lp, flags);
				else out->cdr =NULL;
				break;
			}
		}
		else if(flags & SET_PARSE_QUOTE && *sexpr == '\"')
			instr = 1;	
		else if(flags & SET_PARSE_ESCAPE && *sexpr == '\\')
			inesc = 1;
		else
			exs_nappend(token, sexpr, 1);
		sexpr+=1;
		(*lp)+=1;
	}

	if(strlen(EXS(token))>0)
		_addcar(out, token, flags & SET_PARSE_NIL, flags & SET_PARSE_INSENSITIVE);
	else if(!out->ptr && !out->cdr) {
		free(out);
		out=NULL;
	}

	exs_free(token);
	return out;
}

static int iws(char c)
{
	switch(c)
	{
		case ' ':
		case '\t':
		case '\n':
		case '\r':
			return 1;
		default: return 0;
	}
}

static int nows(const char* str)
{
	register int i=0;
	for(;*str&&iws(*str);str++) i++;
	return !*str?0:i;
}

int se_valid(const char* str)
{
	exString* s = exs_new(str);
	exs_whereip(s, &exspf_only_not, " \t\n\r");

	char *ns = EXS(s);
	int ret = *ns && !(*ns=='(' && memcmp(ns, ns+1, strlen(ns)-1)==0);
	
	exs_free(s);

	return ret;
}

list_t* se_last(list_t *list)
{
	if(list->cdr) return se_last(list->cdr);
	else return list;
}

list_t* se_parse(const char* str, unsigned int flags)
{
	if(!str) return NULL;
	int lp=0,tst=0;
	if(*str=='(' && !(flags & SET_NO_WRAP))
		str+=1;
	else if( (tst=nows(str))>0)
		return se_parse(str+tst, flags);
	if(!se_valid(str)) return NULL;
	return _parse_list(str,&lp, flags);
}

static int _se_add_quotes(const char* inp)
{
	do {
		switch (*inp) {
			case '\"':
			case ' ':
			case ')':
			case '(':
			case '\t':
			case '\r':
				return 1;
			default: break;
		}
		inp+=1;
	}
	while (*inp);
	return 0;
}

void _se_append(exString *out, const char* thing)
{
	int q = _se_add_quotes(thing);
	exString* str = exs_new(NULL);
	
	for(;*thing;thing++)
	{
		switch(*thing)
		{
			case '\n':
			case '\"':
				exs_appendc(str, '\\');
			default:break;
		}
		exs_nappend(str, thing, 1);
	}
	if(q)
	{
		exs_appendc(out, '\"');
		exs_appendc(str, '\"');
	}
	exs_append(out, *str);
	exs_free(str);
}

void _list_tostring(exString *out, const list_t* list)
{
	if(!list->ptr)
		exs_append(out, "nil");
	else if(list->car.sexpr) {
		exs_append(out, "(");
		_list_tostring(out, (list_t*)list->ptr);
	}
	else _se_append(out,(char*)list->ptr);

	if(list->cdr)
	{
		exString *wtf = exs_new(NULL);
		_list_tostring(wtf, list->cdr);
		if(EXS(wtf)[0]!=')')
			exs_append(out, " ");
		exs_append(out, EXS(wtf));
		exs_free(wtf);
	}
	else exs_append(out, ")");
}

void se_print(FILE* fp, const list_t* list)
{
	if(fp == NULL) fp = stdout;
	exString* str = exs_new("(");
	_list_tostring(str, list);
	fprintf(fp, "%s", EXS(str));
	exs_free(str);
}

void se_free(list_t* list)
{
	if(list->cdr) se_free(list->cdr);
	if(list->ptr) {
		if(list->car.alloced) {
			if(list->car.sexpr)
				se_free((list_t*)list->ptr);
			else
				free(list->ptr);
		}
	}
	free(list);
}

enum _se_type se_typeof(list_t* list)
{
	if(!list->ptr) return SE_NIL;
	else if(list->car.sexpr) return SE_LIST;
	else return SE_STRING;
}

static list_t* _se_new(void* car, enum _se_type type, int dupe)
{
	list_t* m =  malloc(sizeof(list_t));
	memset(m,0,sizeof(list_t));
	void* _car = car;
	if(car) {
		switch(type)
		{
			case SE_STRING:
				if(dupe) {
					char* dup = strdup((char*)car);
					m->ptr = dup;
					m->car.size = strlen(dup);
					m->car.alloced = 1;
				}
				else {
					m->ptr = car;
					m->car.size = strlen(car);
				}
				break;
			case SE_LIST:
				if(dupe) { //TODO: Clone alloced cars?
					list_t* cr = malloc(sizeof(list_t));
					list_t* or = cr;
					list_t* car = _car;
					memcpy(cr, car, sizeof(list_t));
					while(car->cdr)	{
						cr->cdr = malloc(sizeof(list_t));
						memcpy(cr->cdr, car->cdr, sizeof(list_t));
						cr->cdr->car.alloced = 0;
						car = car->cdr;
						cr = cr->cdr;
					}
					m->ptr = or;
					m->car.size = sizeof(list_t);
					m->car.alloced = 1;
				}
				else {
					m->ptr = car;
					m->car.size = sizeof(list_t);
				}
				break;
			default: break;
		}
	}
	return m;
}

list_t* se_new(char* str, int dupe)
{
	return _se_new(str, SE_STRING, dupe); //We don't need the list part.
}

car_t se_make(const char* fmt)
{
	if(!fmt) 
	{
		car_t c;
		memset(&c, 0, sizeof(car_t));
		return c;
	}

	while(*fmt && iws(*fmt))
	{
		fmt+=1;
	}
	if(!*fmt) return se_make(NULL);
	else if(*fmt == '(') {
		car_t c;
		list_t* ret =  se_parse(fmt, SETP_DEFAULT);
		memset(&c,0,sizeof(car_t));
		c.ptr = ret;
		c.alloced= 1;
		c.size = sizeof(list_t);
		c.sexpr = 1;
		return c;
	}
	else {
		car_t c;
		memset(&c,0,sizeof(car_t));
		c.ptr = strdup(fmt);
		c.alloced=1;
		c.size = strlen((char*)c.ptr);
		return c;
	}
}

void se_rplaca(list_t* cons, car_t car)
{
	if(cons->car.alloced && cons->ptr)
	{
		switch(se_typeof(cons))
		{
			case SE_LIST:
				se_free(cons->ptr); break;
			case SE_STRING:
				free(cons->ptr); break;
			default: break;
		}
	}
	memcpy(&cons->car, &car, sizeof(car_t));
}

void se_rplacd(list_t* cons, list_t* cdr)
{
	if(cons->cdr)
		se_free(cons->cdr);
	cons->cdr = cdr;
}
